@@ -211,18 +211,13 @@ USER-COMMENTS should be a list of Vorbis comments according to
 `listen-info--opus-comment-header-bindat-spec' or
 `listen-info--flac-comment-block-bindat-spec'.
 
-Return comments in a list of (FIELD . VALUE) cons cells.  Only
-FIELDs that are listed in
-`listen-info--accepted-vorbis-fields' are returned."
+Return comments in a list of (FIELD . VALUE) cons cells."
   (let (comments)
     (dolist (user-comment user-comments)
       (let* ((comment (cdr (assoc 'user-comment user-comment)))
              (pair (listen-info--split-vorbis-comment comment)))
         (push pair comments)))
-    (seq-filter (lambda (elt)
-                  (member (car elt)
-                          listen-info--accepted-vorbis-fields))
-                comments)))
+    comments))
 
 (defun listen-info--split-vorbis-comment (comment)
   "Split Vorbis comment to a field-value pair.
@@ -898,8 +893,7 @@ string."
              (let* ((key-val (split-string str (string 0)))
                     (key (downcase (car key-val)))
                     (val (cadr key-val)))
-               (when (rassoc key listen-info--id3v2-frame-to-info)
-                 (cons key val))))))))
+               (cons key val)))))))
 
 (defun listen-info--decode-id3v2-string (bytes)
   "Decode id3v2 text information from BYTES.
